#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "objz80.h"

extern struct z80fileinfo ofi;
extern unsigned char * workspace;
extern int line;

struct relocentry ** relocentries = NULL;
struct globalsym ** symbols = NULL; /* it could be a hash table, iknow */
struct globalsym ** globals = NULL;
struct externsym ** externs = NULL;

int symcount;

void fi_reset(void)
{
    ofi.relocentries = 0;
    ofi.globals = 0;
    ofi.externs = 0;
}

void add_reloc_entry(unsigned short pc)
{
    struct relocentry * new = (struct relocentry*) malloc(sizeof(struct relocentry));
#ifdef DEBUG
    printf("Add reloc entry %d\n",pc);
#endif

    ofi.relocentries ++;
    relocentries = (struct relocentry **) realloc(relocentries, sizeof(struct relocentry*) * ofi.relocentries);
    relocentries[ofi.relocentries - 1] = new;
    new->offset = pc;
}

void add_symbol(unsigned char * name, unsigned short pc)
{
    struct globalsym * new = (struct globalsym*) malloc(sizeof(struct globalsym));
#ifdef DEBUG
    printf("Symbol %s @ %d\n", name, (int) pc);
#endif

    symcount ++;
    symbols = (struct globalsym **) realloc(symbols, sizeof(struct globalsym*) * symcount);
    symbols[symcount - 1] = new;

    new->offset = pc;
    snprintf(new->name, 20, "%s", name);
}

void add_sym_reference(unsigned char * name, unsigned short pc)
{
    struct externsym * new = (struct externsym*) malloc(sizeof(struct externsym));
#ifdef DEBUG
    printf("AddReference %s @ %d\n",name,pc);
#endif
    ofi.externs ++;
    externs = (struct externsym **) realloc(externs, sizeof(struct externsym*) * ofi.externs);
    externs[ofi.externs - 1] = new;

    new->offset = pc;
    new->mode = RELOCMODE_ABSOLUTE;
    snprintf(new->name, 20, "%s", name);
}

void add_relsym_reference(unsigned char * name, unsigned short pc)
{
    struct externsym * new = (struct externsym*) malloc(sizeof(struct externsym));
#ifdef DEBUG
    printf("AddRelReference %s @ %d\n",name,pc);
#endif
    ofi.externs ++;
    externs = (struct externsym **) realloc(externs, sizeof(struct externsym*) * ofi.externs);
    externs[ofi.externs - 1] = new;

    new->offset = pc;
    new->mode = RELOCMODE_RELATIVE;
    snprintf(new->name, 20, "%s", name);
}

void make_global(unsigned char * name)
{
    struct globalsym * gs;
    int i;
    
    for(i=0; i<symcount; i++) if(strcmp((gs = symbols[i])->name, name) == 0) break;
    if(i == symcount)
    {
	fprintf(stderr,"make_global(line %d): No such symbol: %s\n", line, name);
	exit(1);
    }
    symbols[i] = symbols[symcount - 1];
    symcount --;
    if(symcount == 0)
    {
	free(symbols);
	symbols = NULL;
    }
    else
	symbols = (struct globalsym **) realloc(symbols, sizeof(struct globalsym*) * symcount);
    
    ofi.globals ++;
    globals = (struct globalsym **) realloc(globals, sizeof(struct globalsym*) * ofi.globals);
    globals[ofi.globals - 1] = gs;
}

int lookup_symbol(unsigned char * name)
{
    int i;
    struct globalsym * gs;
    
    for(i=0; i<symcount; i++) if(strcmp((gs = symbols[i])->name, name) == 0) break;
    if(i != symcount) return(gs->offset);
    for(i=0; i<ofi.globals; i++) if(strcmp((gs = globals[i])->name, name) == 0) break;
    if(i != ofi.globals) return(gs->offset);
    
    return(-1);
}

void complete_symbol_values(void)
{
    struct externsym * rel;
    int i;
    int offs;
    
    for(i=0; i < ofi.externs; i++)
    {
	int ddr;
	ddr = lookup_symbol(externs[i]->name);
#ifdef DEBUG
	printf("lookupof %s returned %d\n", externs[i]->name, ddr);
#endif
	if(ddr != -1)
	{
	    offs = externs[i]->offset;
/* ENDIANESS */
	    if(externs[i]->mode == RELOCMODE_ABSOLUTE)
	    {
		add_reloc_entry(offs);
 		workspace[offs] = ddr & 0xff;
	        workspace[offs+1] = (ddr >> 8) & 0xff;
	    }
	    else
	    {
		if((ddr - offs - 1) < -127 || (ddr - offs - 1) > 128) 
		{
		    fprintf(stderr,"reloc(): Relative jump too far (offset %d symbol @ %d)\n", offs, ddr);
		    exit(0);
		}
		workspace[offs] = (ddr - offs - 1) & 0xff;
	    }
	    
	    externs[i] = externs[ofi.externs - 1];
	    ofi.externs --;
	    
	    if(ofi.externs)
		externs = (struct externsym **) realloc(externs, sizeof(struct externsym*) * ofi.externs);
	    else
	    {
		free(externs);
		externs = NULL;
		break;
	    }
	    i--;
	}
    }
}

void save(int fd, int len)
{
    int i;
    
    ofi.code_length = len;
    ofi.header = Z80_MAGIC;
	
    if(write(fd, &ofi, sizeof(struct z80fileinfo)) != sizeof(struct z80fileinfo)) perror("write");
    for(i = 0; i < ofi.relocentries; i++) 
	if(write(fd, relocentries[i], sizeof(struct relocentry)) != sizeof(struct relocentry)) perror("write");
    for(i = 0; i < ofi.externs; i++) 
	if(write(fd, externs[i], sizeof(struct externsym)) != sizeof(struct externsym)) perror("write");
    for(i = 0; i < ofi.globals; i++) 
	if(write(fd, globals[i], sizeof(struct globalsym)) != sizeof(struct globalsym)) perror("write");
	
    write(fd, workspace, len); 
}
